from typing import TYPE_CHECKING, Any, Union, Tuple, Optional, Dict, OrderedDict, \
    Generator, Type, List, ContextManager

from psycopg import sql
from typing_extensions import Protocol, runtime_checkable

from .helper import classproperty

if TYPE_CHECKING:
    from psycopg import connection
    from .columns import Column

__all__ = [
    'Dataclass',
    'DatabaseModel'
]


@runtime_checkable
class Dataclass(Protocol):
    __dataclass_fields__: Dict[str, Any]


# Properties gave warnings in PyCharm, this disables checking that inspection
# noinspection PyPropertyDefinition
@runtime_checkable
class DatabaseModel(Dataclass, Protocol):
    __column_definitions__: OrderedDict[str, 'Column']
    __primary_key__: Optional['Column']

    __schema_name__: str
    __table_name__: str

    __instance_cache__: Dict[Any, 'DatabaseModel']

    @classmethod
    def createTable(cls, conn: 'connection.Connection[Any]', *, recreateSchema: bool = False, recreateTable: bool = False, recreateColumns: bool = False) -> None:
        """
        Create a table representing this class.

        :param conn: the connection to use
        :type conn: psycopg.connection.Connection
        :param recreateSchema: if true it will drop the schema before recreating it. This will drop any other tables on the schema too
        :type recreateSchema: bool
        :param recreateTable: if true it will drop the table before recreating it. This will drop any other tables that depend on it
        :type recreateTable: bool
        :param recreateColumns: if true it will recreate any columns before creating the table.
        :type recreateColumns: bool
        """

    @classmethod
    def instantiateAll(cls, conn: 'connection.Connection[Any]', query: Union[str, 'sql.Composable'] = '') -> Tuple['DatabaseModel', ...]:
        """
        Instantiate all models of this type with the given query.

        :param conn: the connection to use
        :type conn: connection.Connection[Any]
        :param query: the additional query to use after the select statement
        :type query: Union[str, sql.Composable]
        :return: a tuple of every model returned from the query
        :rtype: Tuple[DatabaseModel, ...]
        """

    @classmethod
    def instantiateOne(cls, conn: 'connection.Connection[Any]', query: Union[str, 'sql.Composable'] = '') -> 'DatabaseModel':
        """
        Instantiate one model of this type with the given query.

        :param conn: the connection to use
        :type conn: connection.Connection[Any]
        :param query: the additional query to use after the select statement
        :type query: Union[str, sql.Composable]
        :return: a model
        :rtype: DatabaseModel
        """

    @classmethod
    def instantiate(cls, conn: 'connection.Connection[Any]', query: Union[str, 'sql.Composable'] = '') -> Generator['DatabaseModel', None, None]:
        """
        Instantiate each models of this type with the given query as a generator.

        :param conn: the connection to use
        :type conn: connection.Connection[Any]
        :param query: the additional query to use after the select statement
        :type query: Union[str, sql.Composable]
        :return: a tuple of every model returned from the query
        :rtype: Tuple[DatabaseModel, ...]
        """

    @classmethod
    def instantiateFromPrimaryKey(cls, conn: 'connection.Connection[Any]', primaryKey: Any) -> 'DatabaseModel':
        """
        Instantiate a model from a given primary key. The model must have a primary key and the given primaryKeyColumn
        must correspond to a value.

        :param conn: the connection to use
        :type conn: connection.Connection[Any]
        :param primaryKey: the primary key to look up
        :type primaryKey: Any
        :return: the model
        :rtype: DatabaseModel
        """

    def insert(self, conn: 'connection.Connection[Any]', *, doTypeConversion: bool = True) -> 'DatabaseModel':
        """
        Insert this model into the database.

        :param conn: the connection to use
        :type conn: connection.Connection[Any]
        :param doTypeConversion: if true a conversion function will be called on each field
        :type doTypeConversion: bool
        """

    def update(self, conn: 'connection.Connection[Any]', *, doTypeConversion: bool = True) -> 'DatabaseModel':
        """
        Update this model in the database, will replace any model currently in the database with the updated values.
        If there was not a row with the primary key this model has it will raise an error.

        :param conn: the connection to use
        :type conn: connection.Connection[Any]
        :param doTypeConversion: if true a conversion function will be called on each field
        :type doTypeConversion: bool
        """

    def insertOrUpdate(self, conn: 'connection.Connection[Any]', *, doTypeConversion: bool = True) -> 'DatabaseModel':
        """
        Intelligently either updates or inserts this model into the database. If there was not a row with the primary
        key this model has it will insert it.

        :param conn: the connection to use
        :type conn: connection.Connection[Any]
        :param doTypeConversion: if true a conversion function will be called on each field
        :type doTypeConversion: bool
        """

    def delete(self, conn: 'connection.Connection[Any]') -> bool:
        """
        Delete this model from the database and return whether or not it was deleted.

        :param conn: the connection to use
        :type conn: connection.Connection[Any]
        :return: if the model was deleted
        :rtype: bool
        """

    @classmethod
    def getColumn(cls, name: str) -> 'Column':
        """
        Get a Column from a given name from this model type.

        :param name: the name of the column
        :type name: str
        :return: the column with the given name
        :rtype: Column
        """

    @classproperty
    def primaryKeyColumn(cls: Type['DatabaseModel']) -> Optional['Column']:
        """
        Get the primary key for this model or model type.

        :return: either the primary key column or None
        :rtype: Optional[Column]
        """

    @property
    def primaryKey(self) -> Optional[Any]:
        """
        Get this model's primary key value or None if there is no primary key.

        :return: the primary key value
        :rtype: Optional[Any]
        """

    @classproperty
    def schema(cls: Type['DatabaseModel']) -> str:
        """
        Get the schema for this model.

        :return: the schema this model uses
        :rtype: str
        """

    @classproperty
    def table(cls: Type['DatabaseModel']) -> str:
        """
        Get the table name for this model.

        :return: the table name this model uses
        :rtype: str
        """

    @classproperty
    def columns(cls: Type['DatabaseModel']) -> List['Column']:
        """
        Get all columns for this model.

        :return: a list of columns
        :rtype: List[Column]
        """

    def mutate(self, conn: 'connection.Connection[Any]', updateOnExit: bool) -> ContextManager[None]:
        """
        Enter a context manager which inserts or updates the model at the end of it if no errors occurred

        :param conn: the connection to use
        :type conn: connection.Connection
        :param updateOnExit: wether or not to update upon a successful update
        :type updateOnExit: bool
        """



import datetime
import json
import warnings
from abc import ABC
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Any, Union, Tuple, Optional, Callable, cast, Type

from iso8601 import parse_date
from psycopg import sql

from .representations import FixedPointValue
from .columns import ColumnType, Column
from .exceptions import PrimaryKeyError, NullValueError, EnumValueError, ArrayLengthWarning
from .helper import acceptNone, splitNestedString
from .protocols import DatabaseModel

if TYPE_CHECKING:
    from psycopg import connection


__all__ = [
    'ForeignKey',
    'Composite',
    'PrimaryKey',
    'Unique',
    'NotNull',
    'Array',

    'EnumType',
    'INTEGER',
    'SERIAL',
    'REAL',
    'TEXT',
    'TIMESTAMP',
    'TIMESTAMP_WITH_TIMEZONE',
    'DATE',
    'TIME',

    'JSON',
    'JSONB',

    'BOOL',
    'VARCHAR',
    'CHAR',
    'NUMERIC',

    'datatypeModifiers',
    'constantDatatypes',
    'customizableDatatypes',

    'AUTO_FILLED',
    'NO_DEFAULT',
]

TABLE_OR_TABLE_COLUMN = Union['DatabaseModel', Tuple['DatabaseModel', str]]


class ForeignKey(ColumnType):
    """
    Defines a column to be a foreign key to a different model.
    """

    def __init__(self, model: 'DatabaseModel', schema: str, table: str, column: 'Column') -> None:
        self.model = model

        self.schema = schema
        self.table = table
        self.column = column
        self._rawType = column.rawType

    def __class_getitem__(cls, key: TABLE_OR_TABLE_COLUMN) -> 'ForeignKey':
        # I really want match statements
        if isinstance(key, DatabaseModel):
            if key.__primary_key__ is None:
                raise PrimaryKeyError(f'{key} contains no primary key')
            return cls(key, key.__schema_name__, key.__table_name__, key.__primary_key__)
        return cls(key[0], key[0].__schema_name__, key[0].__table_name__, key[0].getColumn(key[1]))

    def initializeType(self, conn: 'connection.Connection[Any]', recreate: bool) -> None:
        pass

    @property
    def rawType(self) -> str:
        return self._rawType

    @property
    def typeStatement(self) -> 'sql.Composable':
        return sql.SQL('{} REFERENCES {}.{} ({})').format(
            sql.SQL(self._rawType),
            sql.Identifier(self.schema),
            sql.Identifier(self.table),
            sql.SQL(self.column.name)
        )

    def convertDataFromString(self, conn: 'connection.Connection[Any]', string: Optional[str]) -> Any:
        if self.model.__primary_key__ is None:
            raise PrimaryKeyError(f'{self.model} contains no primary key')

        return self.model.instantiateFromPrimaryKey(conn, string)

    def convertInsertableFromData(self, conn: 'connection.Connection[Any]', data: Any) -> Any:
        if self.model.__primary_key__ is None:
            raise PrimaryKeyError(f'{self.model} contains no primary key')

        # Data will be of type self.model so we can get the primary key of data and return it
        data.insertOrUpdate(conn)
        
        return getattr(data, self.model.__primary_key__.name)

    def __str__(self) -> str:
        return f'{self.rawType} REFERENCES "{self.schema}"."{self.table}" ({self.column.name})'


class Composite(ColumnType):
    """
    Creates a composite postgres type.
    """

    def __init__(self, name: str, fields: Tuple[Tuple[str, 'ColumnType'], ...]) -> None:
        self.name = name
        self.fields = fields

    def __class_getitem__(cls, definition: Tuple[str, Tuple[Tuple[str, 'ColumnType'], ...]]) -> 'Composite':
        name, fields = definition
        return cls(name, fields)

    @property
    def typeStatement(self) -> 'sql.Composable':
        return sql.SQL(self.name)

    @property
    def rawType(self) -> str:
        return self.name

    def initializeType(self, conn: 'connection.Connection[Any]', recreate: bool) -> None:
        with conn.cursor() as cur:
            if recreate:
                dropStatement = sql.SQL('DROP TYPE IF EXISTS {} CASCADE;').format(
                    sql.Identifier(self.name)
                )
                cur.execute(dropStatement)

            cur.execute(
                sql.SQL("""
                DO $$ BEGIN
                    CREATE TYPE {} AS ({});
                EXCEPTION
                    WHEN duplicate_object THEN null;
                END $$;
                """).format(
                    sql.Identifier(self.name),
                    sql.SQL(', ').join([
                        sql.SQL('{} {}').format(
                            sql.Identifier(name),
                            column.typeStatement
                        ) for name, column in self.fields
                    ])
                )
            )

    def convertDataFromString(self, conn: 'connection.Connection[Any]', string: Optional[str]) -> Any:
        columns = (c for _, c in self.fields)

        return tuple(c.convertDataFromString(conn, i) for c, i in zip(columns, splitNestedString(string)))

    def convertInsertableFromData(self, conn: 'connection.Connection[Any]', data: Any) -> Any:
        return data


class ModifiedColumnType(ColumnType, ABC):
    def __init__(self, type: 'ColumnType') -> None:
        self.type = type

    def __class_getitem__(cls, type: 'ColumnType') -> 'ModifiedColumnType':
        return cls(type)

    @property
    def typeStatement(self) -> 'sql.Composable':
        return self.type.typeStatement

    def initializeType(self, conn: 'connection.Connection[Any]', recreate: bool) -> None:
        self.type.initializeType(conn, recreate)

    @property
    def rawType(self) -> str:
        return self.type.rawType

    @property
    def primary(self) -> bool:
        return self.type.primary

    def convertDataFromString(self, conn: 'connection.Connection[Any]', string: Optional[str]) -> Any:
        return self.type.convertDataFromString(conn, string)

    def convertInsertableFromData(self, conn: 'connection.Connection[Any]', data: Any) -> Any:
        return self.type.convertInsertableFromData(conn, data)

    def __str__(self) -> str:
        return str(self.type)


class Array(ModifiedColumnType):
    """
    Turns the given collumn into an array, must be used first in any chain of modified types.
    """

    def __init__(self, type: 'ColumnType', length: Optional[int] = None) -> None:
        super().__init__(type)
        self.length = length

    def __class_getitem__(cls, items: Union['ColumnType', Tuple['ColumnType', int]]) -> 'ModifiedColumnType':
        # I really want match statements
        if type(items) == tuple:
            return cls(*items)

        return cls(cast('ColumnType', items))

    @property
    def typeStatement(self) -> 'sql.Composable':
        if self.length is not None:
            return sql.SQL('{}[' + str(self.length) + ']').format(self.type.typeStatement)
        else:
            return sql.SQL('{}[]').format(self.type.typeStatement)

    def convertDataFromString(self, conn: 'connection.Connection[Any]', string: Optional[str]) -> Any:
        if string is None:
            return None

        items = splitNestedString(string)
        
        if self.length is not None and len(items) != self.length:
            warnings.warn(f'Expected {self.length} items, got {len(items)} ({string})', ArrayLengthWarning)
        
        return list(self.type.convertDataFromString(conn, None if item == 'NULL' else item) for item in items)

    def convertInsertableFromData(self, conn: 'connection.Connection[Any]', data: Any) -> Any:
        if data is None:
            return None
        return list(self.type.convertInsertableFromData(conn, d) for d in data)

    def __str__(self) -> str:
        if self.length is not None:
            return str(self.type) + f'[{self.length}]'
        else:
            return str(self.type) + '[]'


class NotNull(ModifiedColumnType):
    """
    Requires a column can not be null. Should be applied to most columns.
    """

    @property
    def typeStatement(self) -> 'sql.Composable':
        return sql.SQL('{} NOT NULL').format(self.type.typeStatement)

    def convertInsertableFromData(self, conn: 'connection.Connection[Any]', data: Any) -> Any:
        if data is None:
            raise NullValueError('Attempted to fill NOT NULL field with null')

        return super().convertInsertableFromData(conn, data)

    def __str__(self) -> str:
        return str(self.type) + ' NOT NULL'


class PrimaryKey(ModifiedColumnType):
    """
    Defines this column as the primary key for a table. There can only be one defined for each table.
    """

    @property
    def typeStatement(self) -> 'sql.Composable':
        return sql.SQL('{} PRIMARY KEY').format(self.type.typeStatement)

    @property
    def primary(self) -> bool:
        return True

    def __str__(self) -> str:
        return str(self.type) + ' PRIMARY KEY'


class Unique(ModifiedColumnType):
    """
    Requires that each value entered into the database is unique for this field.
    """

    @property
    def typeStatement(self) -> 'sql.Composable':
        return sql.SQL('{} UNIQUE').format(self.type.typeStatement)

    def __str__(self) -> str:
        return str(self.type) + ' UNIQUE'


class LiteralType(ColumnType):
    """
    A basic type. The type name and raw name are the same and there are customizable converter functions.
    """

    def __init__(self, literal: str, converter: Callable[[str], Any], inverse: Callable[[Any], Any]) -> None:
        self.type = literal
        self.converter = acceptNone(converter)
        self.inverse = acceptNone(inverse)

    @property
    def typeStatement(self) -> 'sql.Composable':
        return sql.SQL(self.type)

    def initializeType(self, conn: 'connection.Connection[Any]', recreate: bool) -> None:
        pass

    @property
    def rawType(self) -> str:
        return self.type

    def convertDataFromString(self, conn: 'connection.Connection[Any]', string: Optional[str]) -> Any:
        return self.converter(string)

    def convertInsertableFromData(self, conn: 'connection.Connection[Any]', data: Any) -> Any:
        return self.inverse(data)

    def __str__(self) -> str:
        return self.type


class EnumType(ColumnType):
    """
    A constructed type that can only be one of a few values.
    """

    def __init__(self, enumType: Type[Enum]) -> None:
        self.enumType = enumType
        self.type = enumType.__name__.lower()
        self.enums = tuple(t.name.lower() for t in enumType)

        self._enumConversion = {t.name.lower(): t.name for t in enumType}

    def __class_getitem__(cls, args: Union[Type[Enum], Tuple[str, Tuple[str, ...]]]) -> 'EnumType':
        if type(args) == tuple:
            return cls(cast(Type[Enum], Enum(*args)))
        elif issubclass(cast(Type[Enum], args), Enum):
            return cls(cast(Type[Enum], args))
        else:
            raise TypeError(f'{args} is not an enum or enum definition.')

    @property
    def typeStatement(self) -> 'sql.Composable':
        return sql.SQL(self.type)

    def initializeType(self, conn: 'connection.Connection[Any]', recreate: bool) -> None:
        with conn.cursor() as cur:
            if recreate:
                dropStatement = sql.SQL('DROP TYPE IF EXISTS {} CASCADE;').format(
                    sql.Identifier(self.type)
                )
                cur.execute(dropStatement)

            statement = sql.SQL('''
            DO $$ BEGIN
                CREATE TYPE {} AS ENUM ({});
            EXCEPTION
                WHEN duplicate_object THEN null;
            END $$;
            ''').format(
                sql.Identifier(self.type),
                sql.SQL(', ').join(list(map(sql.Literal, self.enums)))
            )
            cur.execute(statement)

    @property
    def rawType(self) -> str:
        return self.type

    def convertDataFromString(self, conn: 'connection.Connection[Any]', string: Optional[str]) -> Any:
        if string is None:
            return None

        return self.enumType[self._enumConversion[string]]

    def convertInsertableFromData(self, conn: 'connection.Connection[Any]', data: Any) -> Any:
        if data is None:
            return None

        if type(data) == str:
            data = self.enumType[self._enumConversion[data]]

        if not isinstance(data, self.enumType):
            raise EnumValueError(f'Attempted to insert {data} into enum {self.type} which only accepts {self.enums}')

        return data.name.lower()

    def __str__(self) -> str:
        return self.type


class PseudoType(LiteralType):
    """
    A type with which its name is different than its raw name.
    """

    def __init__(self, name: str, rawName: str, converter: Callable[[str], Any], inverse: Callable[[Any], Any]) -> None:
        super().__init__(name, converter, inverse)

        self.rawName = rawName

    @property
    def rawType(self) -> str:
        return self.rawName


INTEGER = LiteralType('INTEGER', int, int)
SERIAL = PseudoType('SERIAL', 'INTEGER', int, int)
REAL = LiteralType('DOUBLE PRECISION', float, float)

TEXT = LiteralType('TEXT', str, str)


TIMESTAMP = LiteralType('TIMESTAMP', parse_date, lambda t: t.isoformat())
TIMESTAMP_WITH_TIMEZONE = LiteralType('TIMESTAMP WITH TIME ZONE', parse_date, lambda t: t.isoformat())
DATE = LiteralType('DATE', datetime.date.fromisoformat, lambda t: t.isoformat())
TIME = LiteralType('TIME', datetime.time.fromisoformat, lambda t: t.isoformat())
# INTERVAL = LiteralType('INTERVAL', str, str)

JSON = LiteralType('JSON', json.loads, json.dumps)
JSONB = LiteralType('JSONB', json.loads, json.dumps)

BOOL = LiteralType('BOOLEAN', lambda s: s == 't', bool)


class VARCHAR(LiteralType):
    def __init__(self, n: int, *, _fromGetItem: bool = False) -> None:
        if not _fromGetItem:
            warnings.warn('Use indexing instead of instantiation for VARCHAR types.', DeprecationWarning, 2)

        assert n > 0

        super().__init__(f'VARCHAR({n})', str, str)

    def __class_getitem__(cls, n: int) -> 'VARCHAR':
        return cls(n, _fromGetItem=True)


class CHAR(LiteralType):
    def __init__(self, n: int, *, _fromGetItem: bool = False) -> None:
        if not _fromGetItem:
            warnings.warn('Use indexing instead of instantiation for CHAR types.', DeprecationWarning, 2)

        assert n > 0

        super().__init__(f'CHAR({n})', str, str)

    def __class_getitem__(cls, n: int) -> 'CHAR':
        return cls(n, _fromGetItem=True)


class NUMERIC(LiteralType):
    def __init__(self, precision: int, scale: int, *, _fromGetItem: bool = False) -> None:
        if not _fromGetItem:
            warnings.warn('Use indexing instead of instantiation for NUMERIC types.', DeprecationWarning, 2)

        assert precision > 0
        assert scale >= 0

        super().__init__(f'NUMERIC({precision}, {scale})', partial(FixedPointValue, precision=precision, scale=scale), str)

    def __class_getitem__(cls, args: Tuple[int, int]) -> 'LiteralType':
        return cls(*args, _fromGetItem=True)


datatypeModifiers = [
    ForeignKey,
    Composite,
    Array,
    NotNull,
    PrimaryKey,
    Unique
]

constantDatatypes = [
    INTEGER,
    SERIAL,
    REAL,
    TEXT,
    TIMESTAMP,
    TIMESTAMP_WITH_TIMEZONE,
    DATE,
    TIME,
    JSON,
    JSONB,
    BOOL
]

customizableDatatypes = [
    VARCHAR,
    CHAR,
    NUMERIC
]


class SentinelValue:
    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return f'SentinelValue({self.name})'

    def __str__(self) -> str:
        return self.name


AUTO_FILLED = SentinelValue('AUTO_FILLED')
NO_DEFAULT = SentinelValue('NO_DEFAULT')



import dataclasses
from collections import OrderedDict as OD
from dataclasses import fields, MISSING
from typing import Callable, Any, List, Type, Optional, OrderedDict, Dict, Generator, cast, Tuple, Union, ContextManager

from psycopg import connection, sql

__all__ = [
    'model',
]

from .datatypes import NO_DEFAULT, AUTO_FILLED
from .columns import Column
from .exceptions import PrimaryKeyError, FieldDefaultValueError
from .protocols import Dataclass, DatabaseModel
from .helper import classproperty


class MutationContext:
    def __init__(self, connection: 'connection.Connection[Any]', model: 'DatabaseModel',
                 insertOrUpdateOnExit: bool) -> None:
        self.connection = connection
        self.model = model
        self.insertOrUpdateOnExit = insertOrUpdateOnExit

        self.data: Dict[str, Any] = {}

    def __enter__(self) -> None:
        for f in fields(self.model):
            self.data[f.name] = getattr(self.model, f.name)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if exc_type is None:
            if self.insertOrUpdateOnExit:
                self.model.insertOrUpdate(self.connection)
        else:
            for k, v in self.data.items():
                setattr(self.model, k, v)


def model(_schema: Optional[str] = None, _table: Optional[str] = None, *, useInstanceCache: bool = True) -> \
        Callable[[Type['Dataclass']], Type['DatabaseModel']]:
    def wrapped(cls: Union[Type['Dataclass'], Type[Any]]) -> Type['DatabaseModel']:
        if not isinstance(cls, Dataclass):
            cls = dataclasses.dataclass(cls)

        if _table is None:
            tableName = cls.__name__.lower()
        else:
            tableName = _table

        if _schema is None:
            schemaName = 'public'
        else:
            schemaName = _schema

        columnDefinitions: OrderedDict[str, 'Column'] = OD()
        _primaryKey: Optional['Column'] = None
        primaryKeyIndex: Optional[int] = None

        argsNames: List[str] = []
        allFieldsName: List[str] = []

        # PyCharm does not recognize cls as a Dataclass despite being type hinted as one
        # noinspection PyDataclass
        for i, field in enumerate(fields(cls)):
            definition = Column.fromField(field)

            allFieldsName.append(field.name)
            if field.default is not AUTO_FILLED:
                argsNames.append(field.name)

            columnDefinitions[definition.name] = definition

            if definition.type.primary:
                if _primaryKey is not None:
                    raise PrimaryKeyError(f'{schemaName}.{tableName} ({cls.__name__}) Has two primary keys defined')
                _primaryKey = definition
                primaryKeyIndex = i

            if not (field.default is MISSING or field.default is NO_DEFAULT or field.default is AUTO_FILLED):
                raise FieldDefaultValueError(f'{field.name} does not declare default type of MISSING, NO_DEFAULT, '
                                             f'or AUTO_FILLED')

        argsString = ', '.join(argsNames)
        settersString = '\n'.join(f'    self.{a} = {a}' for a in argsNames)

        replaceAutofilledString = '    for a in dir(self):\n'\
                                  '        if getattr(self, a) is AUTO_FILLED:\n'\
                                  '            setattr(self, a, None) '

        funcString = f"def __init__(self, {argsString}):\n{settersString}\n{replaceAutofilledString}"

        # mypy doesn't support this yet so have to silence the error
        class WrappedClass(cls):  # type: ignore
            __column_definitions__: OrderedDict[str, 'Column'] = columnDefinitions
            __primary_key__: Optional['Column'] = _primaryKey

            __schema_name__: str = schemaName
            __table_name__: str = tableName

            __instance_cache__: Dict[Any, 'WrappedClass'] = {}

            def _create(self, conn: 'connection.Connection[Any]', record: Tuple[Any, ...]) -> None:
                kwargs = {}

                for kc, v in zip(WrappedClass.__column_definitions__.items(), record):
                    k, c = kc
                    kwargs[k] = c.type.convertDataFromString(conn, v)

                for k, v in kwargs.items():
                    setattr(self, k, v)

            def __str__(self) -> str:
                dictlike = ', '.join(f'{a}={getattr(self, a)}' for a in self.__column_definitions__.keys())
                return f'{self.__schema_name__}.{self.__table_name__}({dictlike})'

            def __repr__(self) -> str:
                return str(self)

            def __dir__(self) -> List[str]:
                return list(set(dir(type(self)) + list(self.__dict__.keys())))

            @classmethod
            def getColumn(cls, name: str) -> 'Column':
                return cls.__column_definitions__[name]

            @classproperty
            def primaryKeyColumn(cls: Type['DatabaseModel']) -> Optional['Column']:
                return cls.__primary_key__

            @property
            def primaryKey(self) -> Optional[Any]:
                primary = self.primaryKeyColumn
                if primary is None:
                    return None
                return getattr(self, primary.name)

            @classproperty
            def schema(cls: Type['DatabaseModel']) -> str:
                return cls.__schema_name__

            @classproperty
            def table(cls: Type['DatabaseModel']) -> str:
                return cls.__table_name__

            @classproperty
            def columns(cls: Type['DatabaseModel']) -> List['Column']:
                return list(cls.__column_definitions__.values())

            @classmethod
            def createTable(cls, conn: 'connection.Connection[Any]', *, recreateSchema: bool = False,
                            recreateTable: bool = False, recreateColumns: bool = False) -> None:
                for defini in cls.columns:
                    defini.initialize(conn, recreateColumns)

                createSchema = sql.SQL(
                    'CREATE SCHEMA IF NOT EXISTS {};'
                ).format(
                    sql.Identifier(schemaName)
                )

                createTable = sql.SQL(
                    'CREATE TABLE IF NOT EXISTS {} ({});'
                ).format(
                    sql.Identifier(schemaName, tableName),
                    sql.SQL(', ').join(
                        [d.columnDefinition for d in cls.columns]
                    )
                )

                with conn.cursor() as cur:
                    if recreateSchema:
                        cur.execute(sql.SQL('DROP SCHEMA IF EXISTS {} CASCADE').format(
                            sql.Identifier(schemaName)
                        ))

                    cur.execute(createSchema)

                    if recreateTable:
                        cur.execute(sql.SQL('DROP TABLE IF EXISTS {} CASCADE').format(
                            sql.Identifier(schemaName, tableName)
                        ))

                    cur.execute(createTable)

            @classmethod
            def instantiateAll(cls, conn: 'connection.Connection[Any]', query: Union[str, 'sql.Composable'] = '') -> Tuple['WrappedClass', ...]:
                return tuple(cls.instantiate(conn, query))

            @classmethod
            def instantiateOne(cls, conn: 'connection.Connection[Any]', query: Union[str, 'sql.Composable'] = '') -> 'WrappedClass':
                return next(cls.instantiate(conn, query))

            @classmethod
            def instantiate(cls, conn: 'connection.Connection[Any]', query: Union[str, 'sql.Composable'] = '') -> \
                    Generator['WrappedClass', None, None]:
                if isinstance(query, sql.Composable):
                    additionalQuery = query
                else:
                    additionalQuery = sql.SQL(query)

                queryStatement = sql.SQL('SELECT ({}) FROM {} {};').format(
                    sql.SQL(', ').join(
                        [sql.Identifier(c.name) for c in cls.columns]
                    ),
                    sql.Identifier(schemaName, tableName),
                    additionalQuery
                )

                with conn.cursor() as cur:
                    cur.execute(queryStatement)

                    record = cur.fetchone()

                    while record is not None:
                        record = record[0]
                        if type(record) != tuple:
                            record = (record,)

                        if cls.__primary_key__ is None:
                            # Abuse duck-typing to get "2 init methods" sort of
                            obj = cls(*argsNames)

                            obj._create(conn, record)

                            record = cur.fetchone()

                            yield obj
                        else:
                            primaryKey = cls.__primary_key__.type.convertDataFromString(conn, record[primaryKeyIndex])

                            if useInstanceCache and primaryKey in cls.__instance_cache__:
                                obj = cls.__instance_cache__[primaryKey]
                            else:
                                obj = cls(*argsNames)

                                obj._create(conn, record)

                            if useInstanceCache:
                                cls.__instance_cache__[primaryKey] = obj

                            record = cur.fetchone()

                            yield obj

            @classmethod
            def instantiateFromPrimaryKey(cls, conn: 'connection.Connection[Any]', primaryKey: Any) -> 'DatabaseModel':
                if cls.__primary_key__ is None:
                    raise PrimaryKeyError(f'Model {cls.__name__} has no primary key to instantiate from')

                if useInstanceCache and primaryKey in cls.__instance_cache__:
                    return cls.__instance_cache__[primaryKey]

                obj = cls.instantiateOne(conn, sql.SQL('WHERE {} = {}').format(
                    sql.Identifier(cls.__primary_key__.name),
                    sql.Literal(primaryKey)
                ))

                if useInstanceCache:
                    cls.__instance_cache__[primaryKey] = obj

                return obj

            def insert(self, conn: 'connection.Connection[Any]', *, doTypeConversion: bool = True) -> 'DatabaseModel':
                if doTypeConversion:
                    data = [c.type.convertInsertableFromData(conn, getattr(self, c.name)) for c in self.columns if c.name in argsNames]
                else:
                    data = [getattr(self, c.name) for c in self.columns if c.name in argsNames]

                allColumns = [sql.Identifier(c.name) for c in self.columns]

                insertStatement = sql.SQL('INSERT INTO {} ({}) VALUES ({}) RETURNING ({});').format(
                    sql.Identifier(schemaName, tableName),
                    sql.SQL(', ').join(
                        [sql.Identifier(c.name) for c in self.columns if c.name in argsNames]
                    ),
                    sql.SQL(', ').join(
                        list(map(sql.Literal, data))
                    ),
                    sql.SQL(', ').join(
                        allColumns
                    )
                )

                with conn.cursor() as cur:
                    cur.execute(insertStatement)

                    # After insertion of this object go back and fill in any defaulted fields
                    record = cast(Tuple[Any], cur.fetchone())[0]

                    if type(record) != tuple:
                        record = (record,)

                    self._create(conn, record)

                    if self.__primary_key__ is not None and useInstanceCache:
                        self.__instance_cache__[self.primaryKey] = self

                return self

            def update(self, conn: 'connection.Connection[Any]', *, doTypeConversion: bool = True) -> 'DatabaseModel':
                primary = self.primaryKeyColumn

                if primary is None:
                    raise PrimaryKeyError('Can not update a database model without a primary key.')

                if doTypeConversion:
                    data = [c.type.convertInsertableFromData(conn, getattr(self, c.name)) for c in self.columns]
                else:
                    data = [getattr(self, c.name) for c in self.columns]

                updateStatement = sql.SQL('UPDATE {} SET ({}) = ({}) WHERE {} = {};').format(
                    sql.Identifier(schemaName, tableName),
                    sql.SQL(', ').join(
                        [sql.Identifier(c.name) for c in self.columns]
                    ),
                    sql.SQL(', ').join(
                        list(map(sql.Literal, data))
                    ),
                    sql.Identifier(primary.name),
                    sql.Literal(self.primaryKey)
                )

                with conn.cursor() as cur:
                    cur.execute(updateStatement)

                return self

            def insertOrUpdate(self, conn: 'connection.Connection[Any]', *, doTypeConversion: bool = True) -> 'DatabaseModel':
                if self.primaryKeyColumn is None:
                    raise PrimaryKeyError('Can not insert/update a database model without a primary key.')

                if self.primaryKey is None:
                    return self.insert(conn, doTypeConversion=doTypeConversion)

                instances = WrappedClass.instantiateAll(conn, sql.SQL('WHERE {} = {}').format(
                    sql.Identifier(self.primaryKeyColumn.name),
                    sql.Literal(self.primaryKey)
                ))

                if len(instances) == 0:
                    return self.insert(conn, doTypeConversion=doTypeConversion)
                else:
                    return self.update(conn, doTypeConversion=doTypeConversion)

            def delete(self, conn: 'connection.Connection[Any]') -> bool:
                if self.primaryKeyColumn is None:
                    raise PrimaryKeyError('Can not delete a database model without a primary key.')

                if self.primaryKey is None:  # Never was in the database so it was "deleted"
                    return True

                deleteStatement = sql.SQL('DELETE FROM {} WHERE {} = {} RETURNING {}').format(
                    sql.Identifier(schemaName, tableName),
                    sql.Identifier(self.primaryKeyColumn.name),
                    sql.Literal(self.primaryKey),
                    sql.Identifier(self.primaryKeyColumn.name)
                )

                with conn.cursor() as cur:
                    cur.execute(deleteStatement)

                    return cur.fetchone() is not None

            def mutate(self, conn: 'connection.Connection[Any]', updateOnExit: bool) -> ContextManager[None]:
                return MutationContext(conn, self, updateOnExit)

        miniLocals: Dict[str, Callable[..., None]] = {}

        # Builds init method
        exec(funcString, {'AUTO_FILLED': AUTO_FILLED}, miniLocals)

        # Ignored because this must be done to set init method properly
        WrappedClass.__init__ = miniLocals['__init__']  # type: ignore

        # Transfer wrapped class data over
        WrappedClass.__module__ = cls.__module__
        WrappedClass.__name__ = cls.__name__
        WrappedClass.__qualname__ = cls.__qualname__
        WrappedClass.__doc__ = cls.__doc__

        return WrappedClass

    return wrapped



import ast
from typing import Callable, Any, Optional, Type, TypeVar, List, Union, cast
from functools import wraps

__all__ = [
    'acceptNone',
    'classproperty',
    'identity',
    'splitNestedString'
]


T = TypeVar('T')


def acceptNone(func: Callable[[Any], Any]) -> Callable[[Optional[Any]], Any]:
    @wraps(func)
    def wrapper(x: Optional[Any]) -> Any:
        if x is None:
            return None

        return func(x)
    return wrapper


def identity(x: T) -> T:
    return x


class classproperty:
    """Much like property except only allows for a gettter and works like a classmethod. No instances needed!"""
    def __init__(self, func: Callable[[Type[Any]], Any]) -> None:
        self.func = func

    def __get__(self, _: Any, owner: Type[Any]) -> Any:
        return self.func(owner)


def splitNestedString(arraystring: Optional[Union[str, List[str]]]) -> List[str]:
    """Parses a string from psycopg array/composite type into a list"""
    # Return empty list for null values
    if arraystring is None:
        return []

    if type(arraystring) == list:
        return arraystring

    arraystring = cast(str, arraystring)

    itemstring = arraystring[1:-1]  # Cut off {}

    escaped = False

    inString = False

    depth = 0

    items = []
    startingIndex = 0

    for i, c in enumerate(itemstring):
        if c == ',' and not inString and depth == 0:
            items.append(itemstring[startingIndex:i])
            startingIndex = i + 1

        if escaped:
            escaped = False
        else:
            if c == '"':
                inString = not inString
            elif c == '\\':
                escaped = True

            # Handle multi-dimensional array strings
            if not inString:
                if c == '{' or c == '(':
                    depth += 1
                elif c == '}' or c == ')':
                    depth -= 1

    items.append(itemstring[startingIndex:])

    for i, item in enumerate(items):
        if item[0] == '"':
            items[i] = ast.literal_eval(item)  # Safely eval strings

    return items



import re
from abc import ABC, abstractmethod
from dataclasses import Field
from typing import TYPE_CHECKING, Optional, Any

from psycopg import sql

if TYPE_CHECKING:
    from psycopg import connection


__all__ = [
    'Column',
    'ColumnType'
]


COLUMN_NAME = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]{0,58}')


class Column:
    def __init__(self, name: str, type: 'ColumnType') -> None:
        self.name = name
        self.type = type

    def initialize(self, conn: 'connection.Connection[Any]', recreateColumns: bool = False) -> None:
        self.type.initializeType(conn, recreateColumns)

    @property
    def columnDefinition(self) -> 'sql.Composable':
        return sql.SQL('{} {}').format(sql.Identifier(self.name), self.type.typeStatement)

    @classmethod
    def fromField(cls, field: 'Field[Any]') -> 'Column':
        assert COLUMN_NAME.match(field.name) is not None, f'{field.name} is not a valid column name'
        assert field.name != 'conn', 'Column name must not be "conn"'

        assert isinstance(field.type, ColumnType), 'Fields must be annotated with a type deriving ColumnType'

        return cls(field.name, field.type)

    @property
    def rawType(self) -> str:
        return self.type.rawType

    def __str__(self) -> str:
        return f'"{self.name}" {self.type}'


class ColumnType(ABC):
    """
    Defines the types of data that is allowed in a column of a model.
    """

    @property
    @abstractmethod
    def typeStatement(self) -> 'sql.Composable':
        ...

    @property
    def primary(self) -> bool:
        return False

    @property
    @abstractmethod
    def rawType(self) -> str:
        ...

    @abstractmethod
    def initializeType(self, conn: 'connection.Connection[Any]', recreate: bool) -> None:
        ...

    @abstractmethod
    def convertDataFromString(self, conn: 'connection.Connection[Any]', string: Optional[str]) -> Any:
        """
        Convert string retrieved from the database to a Python object representation.
        Should be the inverse of convertInsertableFromData.

        :param conn: the connection to use
        :type conn: psycopg.connection.Connection
        :param string: the string to convert
        :type string: Optional[str]
        :return: the Python object
        :rtype: Any
        """

    def convertInsertableFromData(self, conn: 'connection.Connection[Any]', data: Any) -> Any:
        """
        Convert a Python object representation to something insertable by psycopg.
        Should be the inverse of convertDataFromString.

        :param conn: the connection to use
        :type conn: psycopg.connection.Connection
        :param data: the object to convert
        :type data: Any
        :return: the insertable object
        :rtype: Any
        """
        return data


import re
from typing import Optional, overload, Union, cast

from .exceptions import FixedPointOverflowError, FixedPointError, FixedPointUnderflowError


__all__ = [
    'FixedPointValue'
]


VALUE_REGEX = re.compile(r'^(-)?(\d+)(?:\.(\d+))?$')


NUMERIC_VALUE = Union[int, float, 'FixedPointValue']


class FixedPointValue:
    @overload
    def __init__(self, value: str) -> None:
        ...

    @overload
    def __init__(self, value: Union[str, int], precision: int, scale: int) -> None:
        ...

    def __init__(self, value: Union[str, int], precision: Optional[int] = None, scale: Optional[int] = None) -> None:
        nones = (1 if precision is None else 0) + (1 if scale is None else 0)

        if nones == 1:
            raise ValueError('Please call one of the overloaded init methods.')

        if type(value) == str:
            match = VALUE_REGEX.match(value)

            if match is None:
                raise ValueError(rf'Invalid value ({value}) must match ^(\d+)(?:\.(\d+))?$')

            sign = -1 if match.group(1) is not None else 1
            integer = match.group(2)
            decimal = match.group(3)

            if decimal is None:
                decimal = ''

            if precision is None:
                precision = len(integer) + len(decimal)

            if scale is None:
                scale = len(decimal)

            if len(integer + decimal + '0' * (scale - len(decimal))) > precision:
                raise FixedPointError(f'Invalid number for given scale and precision: {value}')

            self._value: int = sign * int(integer + decimal + '0' * (scale - len(decimal)))
        elif type(value) == int:
            self._value: int = value  # type: ignore

        if precision is None or precision < 1:
            raise FixedPointError('Invalid precision')

        if scale is None or scale < 0:
            raise FixedPointError('Invalid scale')

        self.precision = precision
        self.scale = scale

        self._maxValue = 10 ** self.precision
        self._minValue = -self._maxValue

        self._lscale = precision - scale

        self._scaleFactor = 1 / 10 ** scale

    def _compatible(self, other: 'FixedPointValue') -> bool:
        return self.precision == other.precision and self.scale == other.scale

    def __float__(self) -> float:
        return self._value * self._scaleFactor  # type: ignore

    def __str__(self) -> str:
        representation = str(self._value)
        representation = '0' * (self.precision - len(representation)) + representation
        representation = representation[:self._lscale] + '.' + representation[self._lscale:]
        return representation.strip('0.')

    def __eq__(self, other: object) -> bool:
        if type(other) == int or type(other) == float:
            return float(self) == other
        elif isinstance(other, FixedPointValue):
            return str(self) == str(other)
        return False

    def __add__(self, other: NUMERIC_VALUE) -> 'FixedPointValue':
        if type(other) == int or type(other) == float:
            other = FixedPointValue(str(other), self.precision, self.scale)

        other = cast('FixedPointValue', other)

        if not self._compatible(other):
            other = other.changePrecisionAndScale(self.precision, self.scale)

        newValue = self._value + other._value

        if newValue <= self._minValue:
            raise FixedPointUnderflowError(newValue, self._minValue)

        if newValue >= self._maxValue:
            raise FixedPointOverflowError(newValue, self._maxValue)

        return FixedPointValue(newValue, self.precision, self.scale)

    def __sub__(self, other: NUMERIC_VALUE) -> 'FixedPointValue':
        return self + -other

    def __mul__(self, other: NUMERIC_VALUE) -> 'FixedPointValue':
        if type(other) == int or type(other) == float:
            other = FixedPointValue(str(other), self.precision, self.scale)

        other = cast('FixedPointValue', other)

        if not self._compatible(other):
            other = other.changePrecisionAndScale(self.precision, self.scale)

        value = int(round(self._value * other._value * self._scaleFactor))

        return FixedPointValue(value, self.precision, self.scale)

    def __truediv__(self, other: NUMERIC_VALUE) -> 'FixedPointValue':
        return self.divide(other, self.scale + self.precision)

    def __neg__(self) -> 'FixedPointValue':
        return FixedPointValue(-self._value, self.precision, self.scale)

    def divide(self, other: NUMERIC_VALUE, newScale: int) -> 'FixedPointValue':
        if type(other) == int or type(other) == float:
            other = FixedPointValue(str(other), self.precision, self.scale)

        other = cast('FixedPointValue', other)

        dividend = self.changeScale(newScale)

        value = dividend._value // other._value

        scale = newScale - other.scale

        return FixedPointValue(value, dividend.precision, scale).changePrecisionAndScale(self.precision, self.scale)

    def changeScale(self, newScale: int) -> 'FixedPointValue':
        return self.changePrecisionAndScale(self.precision + (newScale - self.scale), newScale)

    def changePrecisionAndScale(self, newPrecision: int, newScale: int) -> 'FixedPointValue':
        if newScale < 0:
            raise FixedPointError(f'Invalid scale {newScale}')

        if newPrecision < 1:
            raise FixedPointError(f'Invalid precision {newPrecision}')

        difference = newScale - self.scale

        newValue = int(round(self._value * 10 ** difference))

        maxValue = 10 ** newPrecision

        if newValue >= maxValue:
            raise FixedPointOverflowError(newValue, maxValue)

        if newValue <= -maxValue:
            raise FixedPointUnderflowError(newValue, -maxValue)

        return FixedPointValue(newValue, newPrecision, newScale)



from dataclasses import dataclass
from typing import Optional


__all__ = [
    'Version'
]


@dataclass(frozen=True, order=True)
class Version:
    major: int
    minor: int
    patch: int

    tag: Optional[str] = None

    def __post_init__(self) -> None:
        if self.major < 0 or self.minor < 0 or self.patch < 0:
            raise ValueError(f'Illegal Version Number ({self.major}.{self.minor}.{self.patch})')

        if self.tag == '':
            raise ValueError(f'Illegal Tag Value ({self.tag})')

    def __str__(self) -> str:
        if self.tag is None:
            return f'{self.major}.{self.minor}.{self.patch}'
        return f'{self.major}.{self.minor}.{self.patch}-{self.tag}'



from setuptools import setup, find_packages
import pathlib

from src.databasemodels import version

here = pathlib.Path(__file__).parent.resolve()

description = (here / 'README.md').read_text(encoding='utf-8')

shortDescription = description.split('\n')[2]

setup(
    name='databasemodels',
    version=str(version),
    description=shortDescription,
    long_description=description,
    long_description_content_type='text/markdown',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    python_requires='>=3.7, <4',
    install_requires=[
        'psycopg',
        'psycopg_binary',
        'iso8601',
        'typing_extensions',
    ],
    author='Hazel Rella',
    author_email='hazelrella11@gmail.com',
    url='https://github.com/HazelTheWitch/DatabaseModels',
    license='MIT',
    package_data={
        'databasemodels': ['py.typed']
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ]
)



__all__ = [
    'DBModelError',
    'PrimaryKeyError',
    'NullValueError',
    'EnumValueError',
    'FieldDefaultValueError',
    'DBModelWarning',
    'ArrayLengthWarning',
    'FixedPointError',
    'FixedPointOverflowError',
    'FixedPointUnderflowError',
]


class DBModelError(Exception):
    ...


class PrimaryKeyError(DBModelError):
    ...


class NullValueError(DBModelError):
    ...


class EnumValueError(DBModelError):
    ...


class FieldDefaultValueError(DBModelError):
    ...


class DBModelWarning(RuntimeWarning):
    ...


class ArrayLengthWarning(DBModelWarning):
    ...


class FixedPointError(RuntimeError):
    ...


class FixedPointOverflowError(FixedPointError):
    def __init__(self, value: int, maxValue: int) -> None:
        super().__init__(f'{value} >= {maxValue}')
        self.value = value
        self.maxValue = maxValue


class FixedPointUnderflowError(FixedPointError):
    def __init__(self, value: int, minValue: int) -> None:
        super().__init__(f'{value} <= {minValue}')
        self.value = value
        self.maxValue = minValue
